# mbrobot_plusV2.py
from microbit import i2c, pin0, pin1, pin2, pin13, pin14, pin15, sleep
import machine
import music
import neopixel

_speedPercent = 50
_powerByteL = 50
_powerByteR = 50
_motorState = bytearray(5)
_powerBytesLUT = bytes(b'\x00\x0f\x10\x10\x11\x12\x12\x13\x13\x14\x15\x15\x16\x16\x17\x18\x18\x19\x19\x1a\x1b\x1b\x1c\x1c\x1d\x1e\x1e\x1f\x1f\x20\x21\x21\x22\x22\x23\x24\x24\x25\x25\x26\x27\x29\x2a\x2b\x2c\x2d\x2e\x2f\x30\x31\x32\x33\x34\x35\x36\x37\x38\x39\x3a\x3b\x3b\x3c\x3e\x3f\x41\x43\x45\x47\x49\x4b\x4d\x50\x52\x55\x58\x5b\x5e\x61\x65\x69\x6d\x71\x75\x7a\x7f\x84\x89\x8f\x95\x9b\xa2\xa9\xb1\xb9\xc1\xca\xd3\xdd\xe8\xf3\xff')

_powerOffset = 0
_powerDifferential = 0
_arcScaling = 0
_servoMinPulse = 25
_servoMaxPulse = 131

_ledState = bytearray(b'\x0B\0\0')
_underglowNP = neopixel.NeoPixel(pin15,4)
_alarmSequence = ['c5:1','r','c5,1','r:3']

_UNCONNECTEDERRORMSG = "Error: Please connect to Maqueen robot and switch it on."

def _setMotors(dirL,powerL, dirR, powerR):
    global _motorState
    _motorState[1] = dirL
    _motorState[2] = powerL
    _motorState[3] = dirR
    _motorState[4] = powerR
    try:
        i2c.write(0x10,_motorState)
    except:
        raise RuntimeError(_UNCONNECTEDERRORMSG)

def _setSingleMotor(side,dir, power):
    global _motorState
    _motorState[1 + side]= dir
    _motorState[2 + side] = power
    try:
        i2c.write(0x10,_motorState)
    except:
        raise RuntimeError(_UNCONNECTEDERRORMSG)

def _getPowerByte(speed,offset):
    if speed <= 0:
        return 0
    elif speed < 40:
        return int(0.6 * speed + 15 + offset)
    elif speed < 60:
        return int(speed + offset)
    elif speed < 100:
        return int(min(1.05546** speed + 34 + offset, 255))
    else:
        return 255

def _getPowerByteLUT(speed,offset):
    return min(_powerBytesLUT[speed] + offset,255)

def _getArcBytes(r):
    rmm = int(r * 100)    
    outerSpeed = _speedPercent
    if outerSpeed < 25:
        outerSpeed = 25
    speedFix = min(abs(outerSpeed - 70),20) / 20
    reducedSpeed = 0
    if rmm > 5: 
        n = outerSpeed * (3 * _arcScaling - outerSpeed - 9 *rmm + 220)
        d = -14 * _arcScaling + outerSpeed - 200 +3 * outerSpeed - 10 * rmm - 290
        reducedSpeed = int(n/d)
        if reducedSpeed < 2:
            reducedSpeed = 2 if rmm > 15 else 1
    innerByte = _getPowerByteLUT(int(reducedSpeed),0)
    outerByte = _getPowerByteLUT(int(outerSpeed),0)
    return (innerByte,outerByte)

def calibrate(offset,
differential=0,arcScaling=0):
    global _powerDifferential
    global _powerOffset
    global _arcScaling
    _powerOffset = max(min(int(offset),50),-14)
    _powerDifferential = max(min(int(differential),150),-150)
    _arcScaling = max(min(arcScaling,50),-50)
    setSpeed(_speedPercent)

def setSpeed(speed):
    global _speedPercent
    global _powerByteL
    global _powerByteR
    _speedPercent = int(min(max(speed,0),100))
    powerByte = _getPowerByteLUT(_speedPercent,_powerOffset)
    boost = round((1 - _speedPercent / 100) * abs(_powerDifferential)) if _speedPercent > 0 else 0
    reduction = round((_speedPercent/ 100)* abs(_powerDifferential))
    if _powerDifferential > 0:
        _powerByteL = powerByte - reduction
        _powerByteR = powerByte + boost
    else:
        _powerByteL = powerByte + boost
        _powerByteR = powerByte - reduction

def stop():
    _setMotors(0,0, 0, 0)    

def forward():
    _setMotors(0,_powerByteL,0, _powerByteR)

def backward():
    _setMotors(1,_powerByteL,1,_powerByteR)          

def left():
    _setMotors(1,_powerByteL,0, _powerByteR)

def right():
    _setMotors(0,_powerByteL,1, _powerByteR)

def rightArc(radius):
    inner,outer = _getArcBytes(radius)
    _setMotors(0,outer, 0, inner)

def leftArc(radius):
    inner,outer = _getArcBytes(radius)
    _setMotors(0,inner, 0, outer)

class Motor:
    def __init__(self,side):
        self._side = side

    def rotate(self, speed):      
        speedClamped = int(min(max(abs(speed),0), 100))
        power = _getPowerByteLUT(speedClamped,_powerOffset)
        direction = 0 if speed > 0 else 1
        _setSingleMotor(self._side,direction, power)

def setServo(servo,angle):
    if servo == "P0" or servo == 'S1':
        pin = pin0
    elif servo == "P1" or servo == 'S2':
        pin = pin1
    elif servo == "P2":
        pin = pin2
    else:
        raise ValueError("Unknown Servo. Please use 'P0', 'P1' or 'P2'.")

    if angle < 0 or angle >180:
        raise ValueError("Invalid angle. Must be between 0 and 180")
    frac = (_servoMaxPulse - _servoMinPulse) * int(angle)
    offset = (frac >> 8)+ (frac >> 10) + (frac >> 11)+ (frac >> 12) # / 180
    usPulseTime = _servoMinPulse + offset 
    pin.set_analog_period(20)
    pin.write_analog(usPulseTime)

def setMinServoDuty(duty):    
    global _servoMinPulse
    _servoMinPulse = int(duty)

def setMaxServoDuty(duty):
    global _servoMaxPulse
    _servoMaxPulse = int(duty)    

class IR:
    _address = bytes(b'\x1D')
    def __init__(self,index):
        self.index = index

    def read_digital(self):
        try:
            i2c.write(0x10,IR._address)
        except:
            raise RuntimeError(_UNCONNECTEDERRORMSG)
        byte = ~i2c.read(0x10,1)[0]
        return (byte & (2 ** self.index)) >> self.index
  
    def read_analog(self):
        try:
            i2c.write(0x10,IR._address)
        except:
            raise RuntimeError(_UNCONNECTEDERRORMSG)
        buffer = i2c.read(0x10,11)
        return buffer[1+ self.index * 2]

def getDistance():
    pin13.write_digital(1)
    pin13.write_digital(0)
    p = machine.time_pulse_us(pin14,1, 50000)
    cm =(p >> 6)+ (p >> 10)+ (p >> 11)+ (p>> 12)+ 1
    return max(min(cm,500),2) if cm > 0 else -1 

def setLED(state,stateR=None):
    global _ledState
    stateR = stateR if stateR != None else state
    _ledState[1]= state
    _ledState[2]= stateR
    i2c.write(0x10,_ledState)

def setLEDLeft(state):
    global _ledState
    _ledState[1]= state
    i2c.write(0x10,_ledState) 

def setLEDRight(state):
    global _ledState
    _ledState[2]= state
    i2c.write(0x10,_ledState) 

def fillUnderglow(red,green,blue):
    for i in range(4):
        _underglowNP[i] = (red,green, blue)
    _underglowNP.show()

def setUnderglow(position,red, green, blue):
    if position < 0 or position > 3:
        raise ValueError("invalid underglow position. Must be 0,1,2 or 3.")
    _underglowNP[position]= (red,green,blue)
    _underglowNP.show()

def setAlarm(state):
    if state:
        music.play(_alarmSequence,wait=False,loop=True)    
    else:
        music.stop()

def beep():
    music.pitch(440,200, wait=False)  

pin2.set_pull(pin2.NO_PULL)
delay = sleep
irR2 = IR(0)
irR1 = IR(1)
irRight = irR1
irM = IR(2)
irL1 = IR(3)
irLeft = irL1
irL2 = IR(4)
motL = Motor(0)
motR =Motor(2)